/*********** disjunct.c ****************/

/* New version started March 23,1998 J. Culberson */

/*
This data structure must support disjunctive propagation on vectors of the sort
<x,y,cx,cy> where x and y are distinct vertices, and either x must receive
color cx or y must receive color cy.

CRITERIA:
(1) For each pair of vertices (x,y) there may be up to two tuples; i.e. color
pairs.

(2) Initialization of disjuncts:
	(a) If (x,y) is an edge, A[x] ^ A[y] not empty, and |Ax| =2 &
	|A[y]| = 2 then the edge (x,y) will induce one or two disjuncts.
	i.e. for c in A[x]^A[y], <x,y,A[x]-{c}, A[y]-{c}>

	(b) If (x,y,z) is a triangle (3-clique), |A[x] U A[y] U A[z]|=3,
	and |A[x]| = 2, then induce <y,z,A[y]-A[x],A[z]-A[x]>
	(we assume forcing sequences and failures have already been eliminated,
	e.g. A[y]-A[z] = empty then cz=A[z]-A[x] is forced, etc.)

We do not currently support other 3-avail initialization conditions.

So that we do not have to recheck edges, all edge forcings will be saved and
only undone on backtrack/failure (restore-all-at-depth).

(3) Color remove, force color, and fail conditions local to a disjunct will be
checked when the disjunct is added to the set. If successful, then the
disjunct will also be added to a pending queue for possible later propagation
of other disjuncts.

(4) Propagation of disjuncts  will only entail propagating on the disjunct
datastructure, since all edge induced disjuncts will either be in the set,
or pending on the avail-2 queue (s2).

(5) For efficiency, the disjunct structure should allow finding/testing
all disjuncts on the pair (x,y) in O(1) time and  for any x, finding
all y such that there exists a disjunct on <x,y> in O(#disjuncts).
Thus, we need a fast data structure, and must be updatable, both insertion and
deletion (stack order).

*/


typedef struct p_struct { /* for each x,y pair record cx, cy disjunct pairs */
	int cx[2], cy[2]; /* store symmetrically */
} pairset;

pairset **disjmatrix; /* matrix of disjunction pairs held globally
			this will be dynamic with stack order of search */

int *disjtopdepth;  /* stack to indicate how many elements on
			the disjunct active set at search depth i */
typedef struct act_struct {
	int x,y; /* vertices activated */
	int which; /* 0 or 1, the set activated */
} activerec;

activerec *activestack; /* stack saying which records in disjmatrix
		are active - disjtopdepth will say how many on stack
		at depth i */
/* NOTE: ADJUNCT STRUCTURE IMPLEMENTED
 For more efficiency, we can use an adjacency list in parallel
 with the active matrix and stacks, based on the integerset implementation,
 to store the pairs that are active at any time. This will speed up
 the propagation step, at some over constant head. However, add later
 as this D.S. is complex to implement.
*/
settype *adjunct;

typedef struct pend_struct { /* records for pending disjunct prop. */
	int x,y,cx,cy;
} pendrec;

pendrec *pendingqueue; /* queue of disjuncts pending propagation */
int pendingfront, pendingend; /* queue delimiters */

void DisjAlloc()
{
/* allocate the various arrays for this system */
/* DO THIS ONCE BEFORE USING */

	int i;
	/* by using calloc all elements are zeroed */

	if (NULL == 
		(disjmatrix = (pairset **) calloc(order,sizeof(pairset *)))) {
		printf("cannot allocate disjmatrix in DisjAlloc\n");
		exit(1);
	}

	for(i=0;i<order;i++) {
	  if (NULL == (disjmatrix[i] = (pairset *)
			calloc(order,sizeof(pairset)))) {
		printf("cannot allocate disjmatrix[%d] in DisjAlloc\n",i);
		exit(1);
	  }
	}

	if (NULL ==
		(disjtopdepth = (int *) calloc(order,sizeof(int)))) {
		printf("cannot allocate disjtopdepth in DisjAlloc\n");
		exit(1);
	}

	if (NULL == (activestack  = (activerec *) 
			calloc(order*order,sizeof(activerec)))) {
		printf("cannot allocate activestack in DisjAlloc\n");
		exit(1);
	}

	if (NULL == (pendingqueue = (pendrec *) 
			calloc(order*order,sizeof(pendrec)))) {
		printf("cannot allocate pendingqueue in DisjAlloc\n");
		exit(1);
	}

	if (NULL == 
		(adjunct = (settype *) calloc(order,sizeof(settype)))){
		printf("Cannot allocate the adjunct array in DisjAlloc\n");
		exit(1);
	}
	for(i=0;i<order;i++) {
		adjunct[i] = MakeEmpty(order);
	}

	pendingfront = 0;
	pendingend = 0;


} /* end DisjAlloc */

void DisjRelease()
/*
      Remove the memory reserved by DisjAlloc
*/
{
      int i;
      for(i=0;i<order;i++) {
              Remove(adjunct[i]);
      }
      free(adjunct);
      free(pendingqueue);
      free(activestack);
      free(disjtopdepth);

      for(i=0;i<order;i++)
              free(disjmatrix[i]);

      free(disjmatrix);
} /* end of DisjRelease */


int IsActive(int x, int y, int cx, int cy)
/*
	Return YES if currently active else NO
*/
{
	if   ( ( (disjmatrix[x][y].cx[0] == cx )
		&& (disjmatrix[x][y].cy[0] == cy ))
	 ||  ( (disjmatrix[x][y].cx[1] == cx )
		&& (disjmatrix[x][y].cy[1] == cy )) )
		return YES;
	return NO;
} /* end IsActive */

int RemoveCommon(adjlist G,
		settype s1,
		settype s2,
		settype s3,
		int x, int y,
		int c,
		int depth,
		int statstype)
/* check all common neighbors of x,y and remove color c from them */
{
	int w;
	w = FirstNbr(G,x);
	while ( w != EndList(G->nbrlist[x]) ) {
		if ( IsNbr(G,y,w) ) {
			if (!RemoveAvail(c,w,s1,s2,s3,depth,statstype) )
				return NO;
			
		}
		w = NextNbr(G,x,w);
	}
	return YES;
} /* end RemoveCommon */

void AddActive(int x, int y, int cx, int cy, int depth)
/*
	add teh tuple to the active set at depth
*/
{
	int k;

	k= 0;
	if (disjmatrix[x][y].cx[0] != 0) 
		k = 1;

	disjmatrix[x][y].cx[k] = cx;
	disjmatrix[x][y].cy[k] = cy;

	/* store symmetrically */
	disjmatrix[y][x].cx[k] = cy;
	disjmatrix[y][x].cy[k] = cx;

	activestack[disjtopdepth[depth]].which = k;
	activestack[disjtopdepth[depth]].x = x;
	activestack[disjtopdepth[depth]].y = y;
	disjtopdepth[depth]++;

	/* update adjunct ds */
	if (k==0) {
		if (Member(adjunct[x],y)) {
			printf("depth = %d x=%d y=%d\n",depth,x,y);
		} else
		AddOne(adjunct[x],y);
		if (Member(adjunct[y],x)) {
			printf("\tdepth = %d x=%d y=%d\n",depth,x,y);
		} else
		AddOne(adjunct[y],x);
	}

} /* end AddActive */

void AddPending(int x, int y, int cx, int cy)
/*
	Add a tuple to the pending queue for furhter propagation
*/
{
	pendingqueue[pendingend].x = x;
	pendingqueue[pendingend].y = y;
	pendingqueue[pendingend].cx = cx;
	pendingqueue[pendingend].cy = cy;
	pendingend++;
} /* end AddPending*/

int AddTuple(adjlist G,settype  s1,settype s2,settype s3,
	int x, int y, int cx, int cy, int depth, int statstype)
/*
	add a tuple to the disjunctive sets if not already present
	Return NO if any contradiction seen, else YES
*/

{
#ifdef DISJUNCTPAIR
	if ((numavail[cx] != 1) || (numavail[cy]!= 1)) {
		/* calls from triangle caught here */
	/*
		This happens when colors are removed from
		a vertex during exectution of OneofTriPairs
		in triangle.c. As a result the subtraction
		passes two colors in cx and/or cy. These will be
		caught later as the vertex with avail 1 will be
		on the s1 set, so ignore here. Hoever, if 
		other propagators are added, may want to turnon
		the following output, and if too many it may
		pay to process these.
	*	printf("x,y,cx,cy,a[x],a[y] = %d %d %d %d %d %d\n",
	*		x,y,cx,cy,avail[x],avail[y]);
	*	printf("\tmatrix %d %d %d %d\n",
	*		disjmatrix[x][y].cx[0],disjmatrix[x][y].cy[0],
	*		disjmatrix[x][y].cx[1],disjmatrix[x][y].cy[1]);
	*	printf("depth = %d\n",depth);
	*/
		return YES;
	}
	/* if the tuple is already present do nothing */
	if (IsActive(x,y,cx,cy) )
		return YES;

	/* (A) third tuple on pair */
	if ( (disjmatrix[x][y].cx[0] != 0) && (disjmatrix[x][y].cx[1] != 0)) {
		/* since two active and this one not active then 3 active
		   on this vertex pair. This forces either a FAIL or
		   forced coloring, in either case no addition tothe
		   active pair. */

		if (  ( cx != disjmatrix[x][y].cx[0] )
		   && ( cx != disjmatrix[x][y].cx[1] )
		   && ( cy != disjmatrix[x][y].cy[0] )
		   && ( cy != disjmatrix[x][y].cy[1] ) )
			return NO;

		if ( (disjmatrix[x][y].cx[0] == cx)
			|| (disjmatrix[x][y].cx[1] == cx) ) {
			/* cx is forced on x */
			if (!RemoveAvail((avail[x]&~cx),
					x,s1,s2,s3,depth,DISJUNCTPAIR))
				return NO;
			/* find the y color that is forced */
			if (cx != disjmatrix[x][y].cx[0]) {
			  if (!RemoveAvail((avail[y]&~disjmatrix[x][y].cy[0]),
					y,s1,s2,s3,depth,DISJUNCTPAIR))
				return NO;
			} else if (cx != disjmatrix[x][y].cx[1]) {
			  if (!RemoveAvail((avail[y]&~disjmatrix[x][y].cy[1]),
					y,s1,s2,s3,depth,DISJUNCTPAIR))
				return NO;
			}
			return YES;
		} 
		if ( (disjmatrix[x][y].cy[0] == cy)
			|| (disjmatrix[x][y].cy[1] == cy) ) {
			/* cy is forced on y */
			if (!RemoveAvail((avail[y]&~cy),
					y,s1,s2,s3,depth,DISJUNCTPAIR))
				return NO;
			/* find the color forced on x */
			if (cy != disjmatrix[x][y].cy[0]) {
			  if (!RemoveAvail((avail[x]&~disjmatrix[x][y].cx[0]),
					x,s1,s2,s3,depth,DISJUNCTPAIR))
				return NO;
			} else if (cy != disjmatrix[x][y].cy[1]) {
			  if (!RemoveAvail((avail[x]&~disjmatrix[x][y].cx[1]),
					x,s1,s2,s3,depth,DISJUNCTPAIR))
				return NO;
			}
			return YES;

		}
			
		/* the above should take care of most cases but... */
		/* it may be that the rem. cases rarely happen, so flag it 
			as possible error */
		printf("CASE (A) 3 ON A PAIR NOT IMPLEMENTED\n");
		printf("(x,y) = (%d,%d)\n",x,y);
		printf("\tcx[0]=%d,cy[0]=%d\n",disjmatrix[x][y].cx[0],disjmatrix[x][y].cy[0]);
		printf("\tcx[1]=%d,cy[1]=%d\n",disjmatrix[x][y].cx[1],disjmatrix[x][y].cy[1]);
		printf("\tcx=%d cy=%d\n",cx,cy);
		printf("\tavail[x] = %d avail[y] =%d\n",avail[x],avail[y]);

		/* ignore for now FIX */
		return YES;

	}

	/* (B) Complementary pair, => MERGE */
	/* 
	if ((disjmatrix[x][y].cx[0] == cy) && (disjmatrix[x][y].cy[0] == cx)) {
		aargh - this may cause prolems with calling procedures
		doing edge list traversals (e.g. in triangle)
		So this merge test will be delayed until DisjProp
		when it can be done safely since DisjProp does not 
		depend on graph structure
	}
	*/

	/* (C) Forced Coloring */
	if (disjmatrix[x][y].cx[0] != 0) {
	 if ((disjmatrix[x][y].cx[0] == cx ) && (disjmatrix[x][y].cy[0] != cy)){
		if (!RemoveAvail((avail[x]&~cx),x,s1,s2,s3,depth,DISJUNCTPAIR))
			return NO;
		return YES;
	 }
	 if ((disjmatrix[x][y].cy[0] == cy ) && (disjmatrix[x][y].cx[0] != cx)){
                if (!RemoveAvail((avail[y]&~cy),y,s1,s2,s3,depth,DISJUNCTPAIR))
                        return NO;
                return YES;
	 }
        }

#endif /* DISJUNCTPAIR */
	/* if cx == cy remove color from common nbrs 
		NOTE: this can be done even if DISJUNCTPAIR turned off */
	if (cx == cy )
		if (!RemoveCommon(G,s1,s2,s3,x,y,cx,depth,statstype) )
			return NO;


#ifdef DISJUNCTPAIR
	/* (D) Add to the active set and to the pending queue */
	AddActive(x,y,cx,cy,depth);
	AddPending(x,y,cx,cy); 
#endif
	return YES;
} /* end AddTuple */

#ifdef DISJUNCTPAIR
int DisjProp(adjlist G,settype  s1,settype s2,settype s3, int depth)
{
#ifdef DISJPROP

	int x,y,cx,cy,c;
	int w;
	int av,v;
	/* get first tuple from pending queue */
	x = pendingqueue[pendingfront].x;
	y = pendingqueue[pendingfront].y;
	cx = pendingqueue[pendingfront].cx;
	cy = pendingqueue[pendingfront].cy;
	pendingfront++;

	/* verify vertices  still in graph  and neither vertex  in s1 
		return YES if fails any of these */
	if ( (!Member(G->vlist,x)) || ( !Member(G->vlist,y)) || 
		(numavail[avail[x]] <2 ) || (numavail[avail[y]] < 2) )
		return YES; 
	
	/* if not active then deactivated soemhow after addition to pending
	 * means this is an error
	*/
	if (!IsActive(x,y,cx,cy)) {
		printf("ERROR: x,y,cx,cy %d %d %d %d not active in DisjProp\n",
			x,y,cx,cy);
		exit(1);
	}

	/* Do a merge check -- we skipped this in AddTuple */
	if ( disjmatrix[x][y].cx[1] != 0 ) { /* otherwise only one color pair*/
		if ( (disjmatrix[x][y].cx[0] == disjmatrix[x][y].cy[1] )
		  && (disjmatrix[x][y].cx[1] == disjmatrix[x][y].cy[0] ) ){
			if (IsNbr(G,x,y))
			{
#ifdef DISJTRACE
			  printf("Disjunct forced equivalent on edge %d,%d ",
				x+1,y+1);
			  PrintAvail(x,targetcolor);
			  printf(" ");
			  PrintAvail(y,targetcolor);
			  printf("\n");
			  fflush(stdout);
#endif
			  return NO;
			}
			c = avail[x] & ~avail[y];
			Mergev(G,x,y,depth);
			if (!RemoveAvail(c,x,s1,s2,s3,depth,DISJPROP))
				return NO;
			/* the following are not redundant */
			/* the first cuts nodes searched and run time by 1.44
			 * on flat200.4.65.b. I also checked that it is
			 * addition to the s2 queue that matters, adding to s3
			 * did not produce as much results */
			
			if (numavail[avail[x]] == 2) {
				if (!Member(s2,x)) AddOne(s2,x);
			}
			/* this one may be redundant */
			else if (numavail[avail[x]] == 3) {
				if (!Member(s3,x)) AddOne(s3,x);
			}
			return YES;
		}
	}	
	/* check every nbr wrt the disjunct set of each vertex and 
		AddTuple any that meet criteria - return NO if
		failure detected */
	/* FOR MORE EFFICIENCCY USING AN ADJUNCT DATA STRUCTURE */
	w = First(adjunct[x]);
	while(w != EndList(adjunct[x]) ) {
	   if (Member(G->vlist,w)) {
		if ((w != y) && (disjmatrix[x][w].cx[0] != 0)) {
			if (disjmatrix[x][w].cx[0] != cx) {
				if (!AddTuple(G,s1,s2,s3,w,y,
				   disjmatrix[x][w].cy[0],cy,depth,DISJPROP))
					return NO;
			}
		}
		if ((w != y) && (disjmatrix[x][w].cx[1] != 0)) {
			if (disjmatrix[x][w].cx[1] != cx) {
				if (!AddTuple(G,s1,s2,s3,w,y,
				   disjmatrix[x][w].cy[1],cy,depth,DISJPROP))
					return NO;
			}
		}
	  }
	  w = Next(adjunct[x],w);
	} /* while not endlist */
	
	w = First(adjunct[y]);
	while(w != EndList(adjunct[y]) ) {
	   if (Member(G->vlist,w)) {
		if ((w != x) && (disjmatrix[y][w].cx[0] != 0)) {
			if (disjmatrix[y][w].cx[0] != cy) {
				if (!AddTuple(G,s1,s2,s3,w,x,
				   disjmatrix[y][w].cy[0],cx,depth,DISJPROP))
					return NO;
			}
		}
		if ((w != x) && (disjmatrix[y][w].cx[1] != 0)) {
			if (disjmatrix[y][w].cx[1] != cy) {
				if (!AddTuple(G,s1,s2,s3,w,x,
				   disjmatrix[y][w].cy[1],cx,depth,DISJPROP))
					return NO;
			}
		}
	  }
	  w = Next(adjunct[y],w);
	}
	
	// MAY 30,1999 
	// The following checks neighbors of the vertex x
	// and tests whether (a) the disjuncts on y force the coloring of y and
	// (b) if no then is another tuple available?
	//
	// See the Smallk paper for a discussion

	w = FirstNbr(G,x);
	while (w != EndList(G->nbrlist[x])) 
	{
	   if (w != y) 
	//	&& (numavail[avail[w]] >2)) // This condition added june 04,99
		// as it seems this may be posting redundant  tuples.
		// but further tests: what if |A[x]| > 2?
		// experiments indicate it might be a bad idea::
		// nodes go up 9usually) by 10% or more while
		// time does not change - BUt I hope to make further use
		// of disjuncts so for now revert to May 30.
	   {
		av = 0;
		v = FirstNbr(G,w);
		while (v != EndList(G->nbrlist[w])) {
		   if (v != x) {
			if (disjmatrix[y][v].cx[0] == cy)
			{
				av |= disjmatrix[y][v].cy[0];
			}
			else if (disjmatrix[y][v].cx[1] == cy)
			{
				av |= disjmatrix[y][v].cy[1];
			}
		   }
		   v = NextNbr(G,w,v);
		}
		// now that we have the colors that coloring y with cy
		// would remove from w, check if all removed => 
		// c is forced to be colored cy
		if (0 == numavail[avail[w] & ~av])
		{
			if (!RemoveAvail(~cy,y,s1,s2,s3,depth,DISJPROP))
				return NO;
		} // otherwise if there is one color then that is a disjunct
		
		 else if  (1 == numavail[avail[w] & ~av])
		{

			if (!AddTuple(G,s1,s2,s3,w,y,
				   avail[w]&~av,cy,depth,DISJPROP))
				return NO;
		}
	    }
	    w = NextNbr(G,x,w);
	}
	// Now repeat for y
	w = FirstNbr(G,y);
	while (w != EndList(G->nbrlist[y])) 
	{
	   if (w != x)
	    // && (numavail[avail[w]] >2)) // This condition added june 04,99
	   {
		av = 0;
		v = FirstNbr(G,w);
		while (v != EndList(G->nbrlist[w])) {
		   if (v != y) {
			if (disjmatrix[x][v].cx[0] == cx)
			{
				av |= disjmatrix[x][v].cy[0];
			}
			else if (disjmatrix[x][v].cx[1] == cx)
			{
				av |= disjmatrix[x][v].cy[1];
			}
		   }
		   v = NextNbr(G,w,v);
		}
		// now that we have the colors that coloring y with cy
		// would remove from w, check if all removed => 
		// c is forced to be colored cy
		if (0 == numavail[avail[w] & ~av])
		{
			if (!RemoveAvail(~cx,x,s1,s2,s3,depth,DISJPROP))
				return NO;
		} // otherwise if there is one color then that is a disjunct
		
		 else if  (1 == numavail[avail[w] & ~av])
		{

			if (!AddTuple(G,s1,s2,s3,w,x,
				   avail[w]&~av,cx,depth,DISJPROP))
				return NO;
		}
	    }
	    w = NextNbr(G,y,w);
	}

	
#endif /* DISJPROP */

#ifndef DISJPROP
	/* if DisjProp not activated, then this prevents spinning on the
	 * queue*/
	pendingfront++;
#endif
	
	return YES;
} /* end DisjProp */

void UndoDisj(int depth)
{
	
	/* unstack and undo each tuple change at this depth */

	int k;

#ifdef  DEBUGDISJ
	printf("top[depth] = %d[%d] pendingend =%d\n",
		disjtopdepth[depth],depth,pendingend);
#endif


	while (disjtopdepth[depth] > disjtopdepth[depth-1]) {
	 disjtopdepth[depth]--;
	 k = disjtopdepth[depth];
	 disjmatrix[activestack[k].x][activestack[k].y].cx[activestack[k].which] = 0;
	 disjmatrix[activestack[k].y][activestack[k].x].cx[activestack[k].which] = 0;
	  /* remove elements from adjunct if this was first element */
	  if (activestack[k].which == 0) {
		Delete(adjunct[activestack[k].x],activestack[k].y);
		Delete(adjunct[activestack[k].y],activestack[k].x);
	  }
	}
} /* end UndoDisj */
#endif /* DISJUNCTPAIR */

/**********end disjunct.c *************/

