/********** LOCAL COPY OF GRAPH ************
	This progam uses a modified adjacency list
	under the assumption that the graph is sparse.
	This local copy will be reduced and restored as the
	backtrack proceeds.
	    +++++++++ Graph Stacks for recording modifications +++++++++
	      +++ edge stack +++
	CreateChangeStack(G) - initialize the graph changes stack 
	RemoveChangeStact(G) - delete the remaining elemeents of the stack

	PushChange(G,x,y,type,depth) - record an operation 
			to the graph at depth type is 
				delv - delete vertex
				dele - delete an edge
				merg - merge two vertices
				adde - add an edge (during merge)
				nop - no operation (stack bottom)

	TopChangeDepth(G) - return the depth associated with the top of 
		stack
	PopChange(G,&x,&y,&type) - return and remove the top change of the stack
	PrintChange(G) - debugging aid to print change stack.

	     +++++++++ Graph creation and deletion +++++++++
	MakeLocalGraph(vperm) - returns a newly constructed copy of
			the global graph --- USES A LOT OF MEMORY
			initial vertex order is set by vperm,
			but when reductions start being made/unmade
			order changes arbitrarily.


	DeleteGraph(G) - free memory allocated to G and its
			associated stacks


	   +++++++++Graph info and traversal +++++++++
	--- Note: for the following care must be taken while traversing
	--- lists that are actively being updated.

	Gorder(G) - return number of vertices in G
	Degree(G,v)	- return degree of vertex v

	Firstv(G) - return first vertex in list of vertices in G
	Nextv(G,v) - return next vertex in G after v.

	FirstNbr(G,v) -  return the first neighbor of v in G
	NextNbr(G,v,w) - return the neighbor following w in nbrlist of v

	IsNbr(G,v,w) - return YES if w is a neighbor of v in G

	     +++++++++ Graph Modification +++++++++
	DeleteEdge(G,x,y,depth) - delete  edge (x,y) and record on stack
			at depth 
	Deletev(G,v,depth) - delete all edges incident to v and then
			delete v recording all at depth
	Mergev(G,x,y,depth) - add edges from x to all v in N[y]\N[x]
			recording those added and then delete y at depth.
	UndoAll(G,depth) - undo all from top of stack recorded at depth.


	     +++++++++ Graph Output (debugging) +++++++++
	PrintLocalGraph(G) - print out the local graph plus the availset
				of each vertex.
************************************/

/* structures for stacks and graph */
enum action { delv, dele, merg, adde, nop };

typedef struct _changerec *changerec;
struct _changerec {
	enum action operation; /* add or delete */
	int x,y;
	int depth;
	changerec next;
};

typedef struct _adjlist {
	settype vlist; /* set of vertices currently in newgraph
			-- arbitrary order */
	settype *nbrlist; /* set of vertices currently adjacent to vertex i */
	changerec stack; /* stack of changes to G */
	
} *adjlist; /* THIS IS THE LOCAL GRAPH TYPE */


	/*********** Graph Stacks for recording modifications *******/
void CreateChangeStack(adjlist G)
{
	if (NULL ==
		(G->stack = (changerec) malloc( sizeof(struct _changerec)))){
		printf("Edge stack allocation failure\n");
		exit(1);
	}
	G->stack->depth = -1; /* dummy stack bottom to make topdepth easy */
	G->stack->operation = nop;
	G->stack->next = NULL;

} /* end CreateChangeStack */

void PushChange(adjlist G, int x,int y, enum action type,int depth)
{
	changerec t;
	if (NULL ==
                (t = (changerec) malloc( sizeof(struct _changerec)))){
                printf("Change stack allocation failure (PushChange)\n");
                exit(1);
        }
	t->x = x;
	t->y = y;
	t->depth = depth;
	t->operation = type;
	t->next = G->stack;
	G->stack = t;
}

void PopChange(adjlist G, int *x, int *y, enum action *type)
{
	changerec t;

	t = G->stack;

	(*x) = t->x;
	(*y) = t->y;
	(*type) = t->operation;

	G->stack = t->next;
	free(t);

} /*end PopChange */

#ifdef INLINE
#define TopChangeDepth(G) ((G)->stack->depth)
#else
int TopChangeDepth(adjlist G)
{
	return (G->stack->depth);
} /* end TopChangeDepth */
#endif
	
void PrintChangeStack(adjlist G)
{
	changerec t;
	printf("\nCHANGE STACK: \n");
	t = G->stack;
	while (t->depth != -1 ) {	
		switch (t->operation) {
			case delv : printf(" Delete V "); break;
			case dele : printf(" Delete E "); break;
			case adde : printf(" Add E    "); break;
			case merg : printf(" Merge    "); break;
			case nop  : printf(" No Oper  "); break;
		}
		printf("%d %d %d \n",t->x, t->y, t->depth);
		t = t->next;
	}
} /* end PrintChangeStack */

void RemoveChangeStack(adjlist G)
{
        int x,y;
        enum action type;
	if (G->stack != NULL) {
        	while( -1 != TopChangeDepth(G) )
                	PopChange(G,&x,&y,&type);
        	free(G->stack);
	} /* else erased by restore and color */
} /*end RemoveChangeStack */

		

	/*********** Graph Creation and Deletion *********/

adjlist MakeLocalGraph( int *vperm )
/*
	Make the local graph copy 
	Uses the global Graph info to construct.
*/
{
	adjlist G;
	int i,j;

	G = (adjlist) malloc(sizeof(struct _adjlist));

	if ( G == NULL) {
		printf("Cannot allocate local graph\n");
		exit(1);
	}

	G->vlist = MakeEmpty(order);

	G->nbrlist = (settype *) calloc(order, sizeof(settype));
	if (G->nbrlist == NULL) {
		printf("Cannot allocate local graph\n");
		exit(1);
	}

	for(i=order-1;i>=0; i--) { /* since sets stacking,
					we reverse the input order */
		AddOne(G->vlist, vperm[i]);
		G->nbrlist[i] = MakeEmpty(order);
		for(j=0;j<order;j++) {
			if (edge(i,j)) {
				AddOne(G->nbrlist[i], j);
			}
		}
	}

	CreateChangeStack(G);
	return G;
} /* end MakeLocalGraph */

void DeleteGraph(adjlist G)
/*
	Free the memory used by G
*/
{
	int i;
	RemoveChangeStack(G);
	for(i=0;i<order;i++)
		Remove(G->nbrlist[i]);

	free(G->nbrlist);
	Remove(G->vlist);
	free(G);
} /* end DeleteGraph */



	/********** Graph Info and Traversal *************/


/* #ifdef INLINE
#define Gorder(G) (G)->vlist->size
#else
*/
int Gorder(adjlist G)
{
	return(Setsize(G->vlist));
} /* end Gorder */
/* #endif

#ifdef INLINE
#else
*/
int Firstv(adjlist G)
{
	return(First(G->vlist));
} /* end Firstv */
/* #endif
*/

#ifdef INLINE
#define Nextv(G,v)  ((int) G->vlist->next[v])
#else
int Nextv(adjlist G, int v)
{
	if ( ! Member(G->vlist,v) ) {
		printf("Nextv: no %d in G\n",v);
		exit(1);
	}
	return(Next(G->vlist,v));
} /* end Nextv */
#endif

#ifdef INLINE
#define Degree(G,v) ((int) G->nbrlist[v]->size)
#else
int Degree(adjlist G,int v)
{
	if ( ! Member(G->vlist,v) ) {
		printf("Trying to get degree of non vertex %d in graph\n",v);
		exit(1);
	}
	return (Setsize(G->nbrlist[v]));
} /* end Degree */
#endif

#ifdef INLINE
#ifdef GLOBALIZE
#define FirstNbr(G,v) ((int) G->nbrlist[v]->next[NONE])
#else /*NOT GLOBALIZE */
#define FirstNbr(G,v) ((int) G->nbrlist[v]->next[G->nbrlist[v]->NONE])
#endif /*GLOBALIZE */
#else /* NOT INLINE */
int FirstNbr(adjlist G, int v)
{
	if ( ! Member(G->vlist, v) ) {
		printf("Trying to find first neighbor of non-vertex %d\n",v);
		abort(1);
	}
	return( First(G->nbrlist[v]));
} /* end FirstNbr */
#endif

#ifdef INLINE
#define NextNbr(G,v,w) ((int) G->nbrlist[v]->next[w])
#else
int NextNbr(adjlist G, int v, int w)
/* 
	Get the neighbor following w in adjacency list of v
*/
{
	if ( ! Member(G->vlist,v) ) {
		printf("Trying to find NextNbr of %d of non-vertex %d\n", w,v);
		exit(1);
	}
	if ( ! Member(G->nbrlist[v],w) ) {
		printf("Trying to find NextNbr when %d not a nbr of %d\n",w,v);
		abort(1);
	}

	return( Next(G->nbrlist[v],w));
} /* end NextNbr */
#endif

#ifdef INLINE
#define IsNbr(G,x,y) ((int) G->nbrlist[x]->inset[y])
#else
int IsNbr(adjlist G,int x,int y)
{
	return ( Member(G->nbrlist[x],y) );
} /* end IsNbr */
#endif

	/***** Graph Modification ********************/
void DeleteEdge(adjlist G, int x, int y, int depth)
{
	if ( ! IsNbr(G,x,y) ) {
		printf("ERROR: Attempt to delete non-existent edge (%d,%d)\n",
				x,y);
		return;
	}
	PushChange(G,x,y,dele,depth);
	Delete(G->nbrlist[x],y);
	Delete(G->nbrlist[y],x);
	edgesremoved++; /* record stats*/
	if (edgesremoved > maxedgesremoved) maxedgesremoved = edgesremoved;
} /* end DeleteEdge */

void Deletev(adjlist G,int v,int depth)
{
	int w;
	while (Degree(G,v) > 0) {
		w = FirstNbr(G,v);
		DeleteEdge(G,v,w,depth);
	}

	PushChange(G,v, -1,delv, depth);

	Delete(G->vlist,v);
	if (Gorder(G) < mingraphsize) mingraphsize = Gorder(G);/*record stats*/
} /* end Deletev */

void Mergev(adjlist G, int x, int y, int depth)
/*
	Delete y and merge edges to v
	Note that when x colored, and we reach this
	point on wrap up coloring, y must receive
	the same color as x.
*/
	
{
	int w;

	if (IsNbr(G,x,y))
	{
		printf("ERROR: Mergev(%d,%d) vertices are neighbors\n",x+1,y+1);
		exit(1);
	}
	while (Degree(G,y) > 0) {
		w = FirstNbr(G,y);
		DeleteEdge(G,y,w,depth);
		if ( (x != w ) && ( ! IsNbr(G,x,w) ) ) { 
			AddOne(G->nbrlist[x],w);
			AddOne(G->nbrlist[w],x);
			edgesremoved--; /* record stats*/
			PushChange(G,x,w,adde,depth);
		}
	}
	
	Delete(G->vlist,y);
	PushChange(G,y,x,merg,depth);
	if (Gorder(G) < mingraphsize) mingraphsize = Gorder(G);/*record stats*/
	if (edgesremoved > maxedgesremoved) maxedgesremoved = edgesremoved;
} /* end Mergev */


void UndoAll(adjlist G, int depth)
{
	int x,y;
	enum action type;

	while (depth == TopChangeDepth(G) ) {
		PopChange(G,&x,&y,&type);
		switch (type) {
			case delv : 
				/* restore vertex */
				AddOne(G->vlist,x);
				break;
			case dele :
				/* restore edge */
				AddOne(G->nbrlist[x],y);
				AddOne(G->nbrlist[y],x);
				edgesremoved--; /* record stats*/
				break;
			case adde :
				/* delete added edge */
				Delete(G->nbrlist[x],y);
				Delete(G->nbrlist[y],x);
				edgesremoved++; /* record stats*/
				break;
			case merg :
				/* restore vertex */
				AddOne(G->vlist,x);
				break;
			case nop :
				break;
			default :
			     printf("ERROR UndoAll: bad op type from stack\n");
			     exit(1);
		}
	}
	if (Gorder(G) > maxgraphsize) maxgraphsize = Gorder(G);
	if (edgesremoved < minedgesremoved) minedgesremoved = edgesremoved;
	if (edgesremoved > maxedgesremoved) maxedgesremoved = edgesremoved;
} /* end UndoAll */
	

	/**** graph output for debugging purposes *****/

void PrintLocalGraph(adjlist G, int targetcolor)
{
	int j,k,totdegree;

	totdegree =0;
	printf("Current Local Graph has %d vertices\n",Setsize(G->vlist));

	printf("Vertex\tDegree\tColors\t  Neighbors\n");

	k = First(G->vlist);

	while(k != EndList(G->vlist) ) {
		printf("%6d\t%6d\t",k,Degree(G,k));
		totdegree += Degree(G,k);
		PrintAvail(k, targetcolor);
		printf("\t: ");
		j = FirstNbr(G,k);

		while ( j != EndList(G->nbrlist[k]) ) {
			printf("%3d ",j);
			j = NextNbr(G,k,j);
		}
		printf("\n");
		k = Next(G->vlist,k);
	}
	printf("Number of Edges =%d\n",totdegree/2);
	printf("\t---\n");
} /* end PrintLocalGraph */

/************END OF LOCAL COPY ************************/


