/************* TRIANGLE PROPAGATION CODE ***************/
/*
	This code takes care of looking for triangles whose
	union of avails has only 3 colors, and then noting
	any forced colors that can be determined from them
	It may eventually pass pairs to the oneof queue
*/


#ifdef PROPTHREE

int TriMerge( adjlist G,
		int x, int y, int z, /* the triangle */
		int av, /* the available tri-colors */
		settype s1,
		settype s2,
		settype s3, /* for further propagation */
		int depth)
/*
	This procedure looks for vertices adjacent to all three vertices
	and removes three colors, and then tries to merge 
	vertices that form triangles on opposite sides of edges
*/
{
	int w,tempw,mflag;
#ifdef TRITRACE
	printf("Triangle %d %d %d\n",x+1,y+1,z+1);
	fflush(stdout);
#endif

	w = FirstNbr(G,x);
	while(w != EndList(G->nbrlist[x]) ){
		mflag = 0;
		if ( (IsNbr(G,y,w)) && (IsNbr(G,z,w) ) ) {
			/* all three neigbors so none of av can be used
			   on w (4-clique) */
			if ( 0 != (av&avail[w]) ){
				fulltri++; /* record stats */
				if (!RemoveAvail(av,w,s1,s2,s3,depth,PROPTHREE))
					return NO;
			}
		} else if ((w != z) && (IsNbr(G,y,w))) { /*neighbor of x and y*/
			if (av == avail[w]) {
				/* same avails --> merge w TO z */
				trimerge++; /* record stats */
				/* remove colors not in intersection */
			 	if (!RemoveAvail(~avail[w],
						z,s1,s2,s3,depth,PROPTHREE)) 
					return NO;
				tempw = NextNbr(G,x,w);
				Mergev(G,z,w,depth);
				w = tempw;
				mflag = 1;
			}
		} else if ((w != y) && (IsNbr(G,z,w))) { /*neighbor of x and z*/
			if (av == avail[w]) {
				/* same avails --> merge w TO y */
				trimerge++; /* record stats */
				/* remove colors not in intersection */
			 	if (!RemoveAvail(~avail[w],
						y,s1,s2,s3,depth,PROPTHREE)) 
					return NO;
				tempw = NextNbr(G,x,w);
				Mergev(G,y,w,depth);
				w = tempw;
				mflag = 1;
			}
		}
		if (! mflag) w = NextNbr(G,x,w);
	}
	/* Now test the yz edge for triangles opposite x */
	w = FirstNbr(G,y);
	while (w != EndList(G->nbrlist[y]) ) {
		mflag = 0;
		if ((w != x) &&(IsNbr(G,z,w)) && (!IsNbr(G,x,w)))  { /* neighbor of y and z */
			if (av == avail[w]) {
				/* same avails --> merge w TO x */
				trimerge++; /* record stats */
				/* remove colors not in intersection */
			 	if (!RemoveAvail(~avail[w],
						x,s1,s2,s3,depth,PROPTHREE)) 
					return NO;
				tempw = NextNbr(G,y,w);
				Mergev(G,x,w,depth);
				w = tempw;
				mflag = 1;
			}
		}
		if (!mflag) w = NextNbr(G,y,w);
	}
			
	return YES;
} /* end TriMerge */

#ifdef DISJUNCTPAIR
int OneofTriPairs(	adjlist G,
			settype s1,
			settype s2,
			settype s3,
			int x, int y, int z, /* the triangle */
			int depth)
/*
	AddTuple may fail due to immediate forced contradiction 
*/
{
	if ( (numavail[avail[x]] < 2)
	   || (numavail[avail[y]] < 2)
	   || (numavail[avail[z]] < 2) ) 
		return YES; /* leave for the forced coloring  to do */

	if ((numavail[avail[x]] == 2) 
		&& (numavail[avail[y]&~avail[x]] == 1)
		&& (numavail[avail[z]&~avail[x]] == 1))
	{
		if (!AddTuple(G,s1,s2,s3,
			y,z,(avail[y]&~avail[x]),
			(avail[z]&~avail[x]),depth,PROPTHREE))
			return NO;
	}

	if ((numavail[avail[y]] == 2) 
		&& (numavail[avail[x]&~avail[y]] == 1)
		&& (numavail[avail[z]&~avail[y]] == 1))
	{
		if (!AddTuple(G,s1,s2,s3,
			x,z,(avail[x]&~avail[y]),
			(avail[z]&~avail[y]),depth,PROPTHREE))
			return NO;
	}

	if ((numavail[avail[z]] == 2) 
		&& (numavail[avail[x]&~avail[z]] == 1)
		&& (numavail[avail[y]&~avail[z]] == 1))
	{
		if (!AddTuple(G,s1,s2,s3,
			x,y,(avail[x]&~avail[z]),
			(avail[y]&~avail[z]),depth,PROPTHREE))
			return NO;
	}

	return YES;
} /* end OneofTriPairs */
#endif

int triangle(adjlist G, int w, settype s1, settype s2, settype s3, int depth)
/*
	"Modified June4,199" in this version triangles are tested for
	propagation as long as each vertex has at most three colors.
	--
	check for triangles on w involving at most three colors
	(which implies three since triangles of two colors handled
	within propdblnbrs)
	and for each such try to propagate the coloring info
	return NO if color violation else return YES when complete.
*/
{
	int x,y; 
	int av3;
/*
	NOTE: for this to work I must be sure not to collapse
	the vertices in the nbrs of w away -- merge into
	these when appropriate to merge -- unit colorings not propagated until
	after this routine done.. 
*/
	/* it is possible for a vertex to be removed from graph
	   and not this queue due to merging where the avail 
	   set not changed 
	*/
	if (!Member(G->vlist,w)) {
		return YES;
	}

	x = FirstNbr(G,w);
	
	while (x != EndList(G->nbrlist[w]) ) {
		if ( (! Member(s3,x) )  /* to avoid repetition
					if member then get this triangle
					  when x is tested later */
		  && (numavail[avail[x]]<=3) ){
			y = NextNbr(G,w,x);
			while (y != EndList(G->nbrlist[w]) ) {
				if ((!Member(s3,y))  /* if member wait till
						  	later*/
				 && (IsNbr(G,x,y) ) /* forms a triangle */
				 && (numavail[avail[y]] <=3) ) {
				   av3 = avail[w]|avail[x]|avail[y];
				   if (numavail[av3] <= 3)
				    if (!TriMerge(G,w,x,y,av3,s1,s2,s3,depth)) {
						return NO;
				    }
#ifdef DISJUNCTPAIR
				   if (!OneofTriPairs(G,s1,s2,s3, w,x,y,depth))
					return NO;
#endif
				}
				y = NextNbr(G,w,y);
			}
		}
		x = NextNbr(G,w,x);
	}
	return YES;
}/* end triangle */
#endif


/*******************END TRIANGLE PROPAGATION CODE **********/
